@@ -1,5 +1,5 @@
 /*
-  $Id: interface.c,v 1.22 2005/02/10 01:59:06 rocky Exp $
+  $Id: interface.c,v 1.24 2005/04/30 07:15:51 rocky Exp $
 
   Copyright (C) 2005 Rocky Bernstein <rocky@panix.com>
   Copyright (C) 1998 Monty xiphmont@mit.edu
@@ -121,11 +121,7 @@ cdio_cddap_open(cdrom_drive_t *d)
   /*  d->select_speed(d,d->maxspeed); most drives are full speed by default */
 
   if ( -1 == d->bigendianp ) {
-    if (yep == cdio_have_atapi(d->p_cdio))
-      /* Is this right? Might we also have to check for SCSI ATAPI? */
-      d->bigendianp = 1;
-    else
-      d->bigendianp = data_bigendianp(d);
+    d->bigendianp = data_bigendianp(d);
   }
   
 
