@@ -1,5 +1,5 @@
 /*
-  $Id: scan_devices.c,v 1.26 2005/02/07 03:36:02 rocky Exp $
+  $Id: scan_devices.c,v 1.28 2005/04/30 07:15:51 rocky Exp $
 
   Copyright (C) 2004, 2005 Rocky Bernstein <rocky@panix.com>
   Copyright (C) 1998 Monty xiphmont@mit.edu
@@ -292,7 +292,7 @@ cdda_identify_device_cdio(CdIo_t *p_cdio, const char *psz_device,
   d->cdda_device_name = strdup(psz_device);
   d->drive_type       = drive_type;
   d->bigendianp       = -1; /* We don't know yet... */
-  d->nsectors         = -1;
+  d->nsectors         = -1; /* We don't know yet... */
   d->b_swap_bytes     = true;
   
   {
