@@ -1,7 +1,7 @@
 /*
-    $Id: win32_ioctl.c,v 1.8 2005/01/27 11:08:55 rocky Exp $
+    $Id: win32_ioctl.c,v 1.27 2005/03/21 08:33:36 rocky Exp $
 
-    Copyright (C) 2004 Rocky Bernstein <rocky@panix.com>
+    Copyright (C) 2004, 2005 Rocky Bernstein <rocky@panix.com>
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -26,7 +26,7 @@
 # include "config.h"
 #endif
 
-static const char _rcsid[] = "$Id: win32_ioctl.c,v 1.8 2005/01/27 11:08:55 rocky Exp $";
+static const char _rcsid[] = "$Id: win32_ioctl.c,v 1.27 2005/03/21 08:33:36 rocky Exp $";
 
 #ifdef HAVE_WIN32_CDROM
 
@@ -39,7 +39,7 @@ static const char _rcsid[] = "$Id: win32_ioctl.c,v 1.8 2005/01/27 11:08:55 rocky
    sprintf(psz_msg, "error file %s: line %d (%s) %d\n", 
 	   _FILE__, __LINE__, __PRETTY_FUNCTION__, i_err)
 #else
-#include <ddk/ntddstor.h>
+#include <ddk/ntddcdrm.h>
 #include <ddk/ntddscsi.h>
 #include <ddk/scsi.h>
 #define FORMAT_ERROR(i_err, psz_msg) \
@@ -60,56 +60,12 @@ static const char _rcsid[] = "$Id: win32_ioctl.c,v 1.8 2005/01/27 11:08:55 rocky
 
 #include <cdio/cdio.h>
 #include <cdio/sector.h>
+#include <cdio/util.h>
 #include "cdio_assert.h"
-#include <cdio/scsi_mmc.h>
+#include <cdio/mmc.h>
 #include "cdtext_private.h"
 #include "cdio/logging.h"
 
-/* Win32 DeviceIoControl specifics */
-/***** FIXME: #include ntddcdrm.h from Wine, but probably need to 
-   modify it a little.
-*/
-
-#ifndef IOCTL_CDROM_BASE
-#    define IOCTL_CDROM_BASE FILE_DEVICE_CD_ROM
-#endif
-#ifndef IOCTL_CDROM_READ_TOC
-#define IOCTL_CDROM_READ_TOC \
-  CTL_CODE(IOCTL_CDROM_BASE, 0x0000, METHOD_BUFFERED, FILE_READ_ACCESS)
-#endif
-#ifndef IOCTL_CDROM_RAW_READ
-#define IOCTL_CDROM_RAW_READ CTL_CODE(IOCTL_CDROM_BASE, 0x000F, \
-                                      METHOD_OUT_DIRECT, FILE_READ_ACCESS)
-#endif
-
-#ifndef IOCTL_CDROM_READ_Q_CHANNEL
-#define IOCTL_CDROM_READ_Q_CHANNEL \
-  CTL_CODE(IOCTL_CDROM_BASE, 0x000B, METHOD_BUFFERED, FILE_READ_ACCESS)
-#endif
-
-typedef struct {
-   SCSI_PASS_THROUGH Spt;
-   ULONG Filler;
-   UCHAR SenseBuf[32];
-   UCHAR DataBuf[512];
-} SCSI_PASS_THROUGH_WITH_BUFFERS;
-
-typedef struct _TRACK_DATA {
-    UCHAR Format;
-    UCHAR Control : 4;
-    UCHAR Adr : 4;
-    UCHAR TrackNumber;
-    UCHAR Reserved1;
-    UCHAR Address[4];
-} TRACK_DATA, *PTRACK_DATA;
-
-typedef struct _CDROM_TOC {
-    UCHAR Length[2];
-    UCHAR FirstTrack;
-    UCHAR LastTrack;
-    TRACK_DATA TrackData[CDIO_CD_MAX_TRACKS+1];
-} CDROM_TOC, *PCDROM_TOC;
-
 typedef struct _TRACK_DATA_FULL {
     UCHAR SessionNumber;
     UCHAR Control : 4;
@@ -132,42 +88,304 @@ typedef struct _CDROM_TOC_FULL {
     TRACK_DATA_FULL TrackData[CDIO_CD_MAX_TRACKS+3];
 } CDROM_TOC_FULL, *PCDROM_TOC_FULL;
 
-typedef enum _TRACK_MODE_TYPE {
-    YellowMode2,
-    XAForm2,
-    CDDA
-} TRACK_MODE_TYPE, *PTRACK_MODE_TYPE;
-
-typedef struct __RAW_READ_INFO {
-    LARGE_INTEGER DiskOffset;
-    ULONG SectorCount;
-    TRACK_MODE_TYPE TrackMode;
-} RAW_READ_INFO, *PRAW_READ_INFO;
-
-typedef struct _CDROM_SUB_Q_DATA_FORMAT {
-    UCHAR               Format;
-    UCHAR               Track;
-} CDROM_SUB_Q_DATA_FORMAT, *PCDROM_SUB_Q_DATA_FORMAT;
-
-typedef struct _SUB_Q_HEADER {
-  UCHAR Reserved;
-  UCHAR AudioStatus;
-  UCHAR DataLength[2];
-} SUB_Q_HEADER, *PSUB_Q_HEADER;
-
-typedef struct _SUB_Q_MEDIA_CATALOG_NUMBER {
-  SUB_Q_HEADER Header;
-  UCHAR FormatCode;
-  UCHAR Reserved[3];
-  UCHAR Reserved1 : 7;
-  UCHAR Mcval :1;
-  UCHAR MediaCatalog[15];
-} SUB_Q_MEDIA_CATALOG_NUMBER, *PSUB_Q_MEDIA_CATALOG_NUMBER;
+typedef struct {
+   SCSI_PASS_THROUGH Spt;
+   ULONG Filler;
+   UCHAR SenseBuf[32];
+   UCHAR DataBuf[512];
+} SCSI_PASS_THROUGH_WITH_BUFFERS;
 
 #include "win32.h"
 
 #define OP_TIMEOUT_MS 60 
 
+/*!
+  Pause playing CD through analog output
+  
+  @param p_cdio the CD object to be acted upon.
+*/
+driver_return_code_t
+audio_pause_win32ioctl (void *p_user_data)
+{
+  const _img_private_t *p_env = p_user_data;
+  DWORD dw_bytes_returned;
+  
+  bool b_success = 
+    DeviceIoControl(p_env->h_device_handle, IOCTL_CDROM_PAUSE_AUDIO,
+		    NULL, (DWORD) 0, NULL, 0, &dw_bytes_returned, NULL);
+
+  if ( ! b_success ) {
+    char *psz_msg = NULL;
+    long int i_err = GetLastError();
+    FORMAT_ERROR(i_err, psz_msg);
+    if (psz_msg) 
+      cdio_info("Error: %s", psz_msg);
+    else 
+      cdio_info("Error: %ld", i_err);
+    LocalFree(psz_msg);
+    return DRIVER_OP_ERROR;
+  }
+  return DRIVER_OP_SUCCESS;
+}
+
+/*!
+  Playing starting at given MSF through analog output
+  
+  @param p_cdio the CD object to be acted upon.
+*/
+driver_return_code_t
+audio_play_msf_win32ioctl (void *p_user_data, msf_t *p_start_msf, 
+			   msf_t *p_end_msf)
+{
+  const _img_private_t *p_env = p_user_data;
+  CDROM_PLAY_AUDIO_MSF play;
+  DWORD dw_bytes_returned;
+
+  play.StartingM = cdio_from_bcd8(p_start_msf->m);
+  play.StartingS = cdio_from_bcd8(p_start_msf->s);
+  play.StartingF = cdio_from_bcd8(p_start_msf->f);
+
+  play.EndingM   = cdio_from_bcd8(p_end_msf->m);
+  play.EndingS   = cdio_from_bcd8(p_end_msf->s);
+  play.EndingF   = cdio_from_bcd8(p_end_msf->f);
+
+  bool b_success = 
+    DeviceIoControl(p_env->h_device_handle, IOCTL_CDROM_PLAY_AUDIO_MSF,
+		    &play, sizeof(play), NULL, 0, &dw_bytes_returned, NULL);
+  
+  if ( ! b_success ) {
+    char *psz_msg = NULL;
+    long int i_err = GetLastError();
+    FORMAT_ERROR(i_err, psz_msg);
+    if (psz_msg) 
+      cdio_info("Error: %s", psz_msg);
+    else 
+      cdio_info("Error: %ld", i_err);
+    LocalFree(psz_msg);
+    return DRIVER_OP_ERROR;
+  }
+  return DRIVER_OP_SUCCESS;
+  
+}
+
+/*!
+  Read Audio Subchannel information
+  
+  @param p_cdio the CD object to be acted upon.
+  
+*/
+driver_return_code_t
+audio_read_subchannel_win32ioctl (void *p_user_data, 
+				  cdio_subchannel_t *p_subchannel)
+{
+  const _img_private_t *p_env = p_user_data;
+  DWORD dw_bytes_returned;
+  CDROM_SUB_Q_DATA_FORMAT q_data_format;
+  SUB_Q_CHANNEL_DATA q_subchannel_data;
+  
+  q_data_format.Format = CDIO_SUBCHANNEL_CURRENT_POSITION;
+  q_data_format.Track=0; /* Not sure if this has to be set or if so what
+			    it should be. */
+  
+  if( ! DeviceIoControl( p_env->h_device_handle,
+		       IOCTL_CDROM_READ_Q_CHANNEL,
+		       &q_data_format, sizeof(q_data_format), 
+		       &q_subchannel_data, sizeof(q_subchannel_data),
+		       &dw_bytes_returned, NULL ) ) {
+    char *psz_msg = NULL;
+    long int i_err = GetLastError();
+    FORMAT_ERROR(i_err, psz_msg);
+    if (psz_msg) 
+      cdio_info("Error: %s", psz_msg);
+    else 
+      cdio_info("Error: %ld", i_err);
+    LocalFree(psz_msg);
+    return DRIVER_OP_ERROR;
+  }
+  p_subchannel->audio_status = 
+    q_subchannel_data.CurrentPosition.Header.AudioStatus;
+  p_subchannel->track = 
+    q_subchannel_data.CurrentPosition.TrackNumber;
+  p_subchannel->index = 
+    q_subchannel_data.CurrentPosition.IndexNumber;
+  p_subchannel->index = 
+    q_subchannel_data.CurrentPosition.IndexNumber;
+  p_subchannel->address = q_subchannel_data.CurrentPosition.ADR;
+  p_subchannel->control = q_subchannel_data.CurrentPosition.Control;
+
+  { 
+    const UCHAR *abs_addr = 
+      q_subchannel_data.CurrentPosition.AbsoluteAddress;
+    const UCHAR *rel_addr = 
+      q_subchannel_data.CurrentPosition.TrackRelativeAddress;
+
+    p_subchannel->abs_addr.m = cdio_to_bcd8(abs_addr[1]);
+    p_subchannel->abs_addr.s = cdio_to_bcd8(abs_addr[2]);
+    p_subchannel->abs_addr.f = cdio_to_bcd8(abs_addr[3]);
+    p_subchannel->rel_addr.m = cdio_to_bcd8(rel_addr[1]);
+    p_subchannel->rel_addr.s = cdio_to_bcd8(rel_addr[2]);
+    p_subchannel->rel_addr.f = cdio_to_bcd8(rel_addr[3]);
+  }
+
+  return DRIVER_OP_SUCCESS;
+}
+
+
+/*!
+  Resume playing an audio CD.
+  
+  @param p_user_data the CD object to be acted upon.
+  
+*/
+driver_return_code_t
+audio_resume_win32ioctl (void *p_user_data)
+{
+  const _img_private_t *p_env = p_user_data;
+  DWORD dw_bytes_returned;
+  
+  bool b_success = 
+    DeviceIoControl(p_env->h_device_handle, IOCTL_CDROM_RESUME_AUDIO,
+		    NULL, (DWORD) 0, NULL, 0, &dw_bytes_returned, NULL);
+
+  if ( ! b_success ) {
+    char *psz_msg = NULL;
+    long int i_err = GetLastError();
+    FORMAT_ERROR(i_err, psz_msg);
+    if (psz_msg) 
+      cdio_info("Error: %s", psz_msg);
+    else 
+      cdio_info("Error: %ld", i_err);
+    LocalFree(psz_msg);
+    return DRIVER_OP_ERROR;
+  }
+  return DRIVER_OP_SUCCESS;
+}
+
+/*!
+  Set the volume of an audio CD.
+  
+  @param p_user_data pointer to the CD object to be acted upon.
+  @param p_volume pointer to the volume levels
+  
+*/
+driver_return_code_t
+audio_set_volume_win32ioctl (void *p_user_data, 
+			     /*in*/ cdio_audio_volume_t *p_volume)
+{
+  const _img_private_t *p_env = p_user_data;
+  DWORD dw_bytes_returned;
+  
+  bool b_success = 
+    DeviceIoControl(p_env->h_device_handle, IOCTL_CDROM_SET_VOLUME,
+		    p_volume, (DWORD) sizeof(cdio_audio_volume_t), 
+		    NULL, 0, &dw_bytes_returned, NULL);
+
+  if ( ! b_success ) {
+    char *psz_msg = NULL;
+    long int i_err = GetLastError();
+    FORMAT_ERROR(i_err, psz_msg);
+    if (psz_msg) 
+      cdio_info("Error: %s", psz_msg);
+    else 
+      cdio_info("Error: %ld", i_err);
+    LocalFree(psz_msg);
+    return DRIVER_OP_ERROR;
+  }
+  return DRIVER_OP_SUCCESS;
+}
+
+/*!
+  Stop playing an audio CD.
+  
+  @param p_user_data the CD object to be acted upon.
+  
+*/
+driver_return_code_t 
+audio_stop_win32ioctl (void *p_user_data)
+{
+  const _img_private_t *p_env = p_user_data;
+  DWORD dw_bytes_returned;
+  
+  bool b_success = 
+    DeviceIoControl(p_env->h_device_handle, IOCTL_CDROM_STOP_AUDIO,
+		    NULL, (DWORD) 0, NULL, 0, &dw_bytes_returned, NULL);
+
+  if ( ! b_success ) {
+    char *psz_msg = NULL;
+    long int i_err = GetLastError();
+    FORMAT_ERROR(i_err, psz_msg);
+    if (psz_msg) 
+      cdio_info("Error: %s", psz_msg);
+    else 
+      cdio_info("Error: %ld", i_err);
+    LocalFree(psz_msg);
+    return DRIVER_OP_ERROR;
+  }
+  return DRIVER_OP_SUCCESS;
+}
+
+/*!
+  Close the tray of a CD-ROM
+  
+  @param p_user_data the CD object to be acted upon.
+  
+*/
+driver_return_code_t 
+close_tray_win32ioctl (const char *psz_win32_drive)
+{
+#ifdef WIN32
+  DWORD dw_bytes_returned;
+  DWORD dw_access_flags;
+  
+  OSVERSIONINFO ov;
+  HANDLE h_device_handle;
+  bool b_success;
+  
+  memset(&ov,0,sizeof(OSVERSIONINFO));
+  ov.dwOSVersionInfoSize=sizeof(OSVERSIONINFO);
+  GetVersionEx(&ov);
+  
+  if((ov.dwPlatformId==VER_PLATFORM_WIN32_NT) &&        
+     (ov.dwMajorVersion>4))
+    dw_access_flags = GENERIC_READ|GENERIC_WRITE;  /* add gen write on W2k/XP */
+  else dw_access_flags = GENERIC_READ;
+
+  h_device_handle = CreateFile( psz_win32_drive, 
+				dw_access_flags,
+				FILE_SHARE_READ | FILE_SHARE_WRITE, 
+				NULL, 
+				OPEN_EXISTING,
+				FILE_ATTRIBUTE_NORMAL, 
+				NULL );
+
+  if( h_device_handle == INVALID_HANDLE_VALUE ) {
+    return DRIVER_OP_ERROR;
+  }
+
+  b_success = 
+    DeviceIoControl(h_device_handle, IOCTL_STORAGE_LOAD_MEDIA2,
+		    NULL, (DWORD) 0, NULL, 0, &dw_bytes_returned, NULL);
+  
+
+  CloseHandle(h_device_handle);
+  
+  if ( ! b_success ) {
+    char *psz_msg = NULL;
+    long int i_err = GetLastError();
+    FORMAT_ERROR(i_err, psz_msg);
+    if (psz_msg) 
+      cdio_info("Error: %s", psz_msg);
+    else 
+      cdio_info("Error: %ld", i_err);
+    LocalFree(psz_msg);
+    return DRIVER_OP_ERROR;
+  }
+  return DRIVER_OP_SUCCESS;
+#else 
+  return DRIVER_OP_UNSUPPORTED;
+#endif
+}
+
 /*!
   Run a SCSI MMC command. 
  
@@ -184,16 +402,16 @@ typedef struct _SUB_Q_MEDIA_CATALOG_NUMBER {
   Return 0 if command completed successfully.
  */
 int
-run_scsi_cmd_win32ioctl( void *p_user_data, 
-			 unsigned int i_timeout_ms,
-			 unsigned int i_cdb, const scsi_mmc_cdb_t * p_cdb,
-			 scsi_mmc_direction_t e_direction, 
-			 unsigned int i_buf, /*in/out*/ void *p_buf )
+run_mmc_cmd_win32ioctl( void *p_user_data, 
+			unsigned int i_timeout_ms,
+			unsigned int i_cdb, const mmc_cdb_t * p_cdb,
+			mmc_direction_t e_direction, 
+			unsigned int i_buf, /*in/out*/ void *p_buf )
 {
   const _img_private_t *p_env = p_user_data;
   SCSI_PASS_THROUGH_DIRECT sptd;
-  bool success;
-  DWORD dwBytesReturned;
+  bool b_success;
+  DWORD dw_bytes_returned;
   
   sptd.Length  = sizeof(sptd);
   sptd.PathId  = 0;      /* SCSI card ID will be filled in automatically */
@@ -211,42 +429,47 @@ run_scsi_cmd_win32ioctl( void *p_user_data,
   memcpy(sptd.Cdb, p_cdb, i_cdb);
 
   /* Send the command to drive */
-  success=DeviceIoControl(p_env->h_device_handle,
-			  IOCTL_SCSI_PASS_THROUGH_DIRECT,               
-			  (void *)&sptd, 
-			  (DWORD)sizeof(SCSI_PASS_THROUGH_DIRECT),
-			  NULL, 0,                        
-			  &dwBytesReturned,
-			  NULL);
+  b_success = DeviceIoControl(p_env->h_device_handle,
+			      IOCTL_SCSI_PASS_THROUGH_DIRECT,               
+			      (void *)&sptd, 
+			      (DWORD)sizeof(SCSI_PASS_THROUGH_DIRECT),
+			      NULL, 0,                        
+			      &dw_bytes_returned,
+			      NULL);
 
-  if(! success) {
+  if ( !b_success ) {
     char *psz_msg = NULL;
     long int i_err = GetLastError();
     FORMAT_ERROR(i_err, psz_msg);
-    cdio_info("Error: %s", psz_msg);
+    if (psz_msg) 
+      cdio_info("Error: %s", psz_msg);
+    else 
+      cdio_info("Error: %ld", i_err);
     LocalFree(psz_msg);
-    return 1;
+    return DRIVER_OP_ERROR;
   }
-  return 0;
+  return DRIVER_OP_SUCCESS;
 }
 
 /*! 
   Get disc type associated with cd object.
 */
-discmode_t
-get_discmode_win32ioctl (_img_private_t *p_env)
+static discmode_t
+dvd_discmode_win32ioctl (_img_private_t *p_env)
 {
-  track_t i_track;
   discmode_t discmode=CDIO_DISC_MODE_NO_INFO;
+  driver_return_code_t rc;
 
   /* See if this is a DVD. */
   cdio_dvd_struct_t dvd;  /* DVD READ STRUCT for layer 0. */
 
   dvd.physical.type = CDIO_DVD_STRUCT_PHYSICAL;
   dvd.physical.layer_num = 0;
-  if (0 == scsi_mmc_get_dvd_struct_physical_private (p_env, 
-						     &run_scsi_cmd_win32ioctl,
-						     &dvd)) {
+
+  rc = mmc_get_dvd_struct_physical_private (p_env, &run_mmc_cmd_win32ioctl,
+					    &dvd);
+
+  if (DRIVER_OP_SUCCESS == rc) {
     switch(dvd.physical.layer[0].book_type) {
     case CDIO_DVD_BOOK_DVD_ROM:  return CDIO_DISC_MODE_DVD_ROM;
     case CDIO_DVD_BOOK_DVD_RAM:  return CDIO_DISC_MODE_DVD_RAM;
@@ -257,12 +480,28 @@ get_discmode_win32ioctl (_img_private_t *p_env)
     default: return CDIO_DISC_MODE_DVD_OTHER;
     }
   }
+  return discmode;
+}
 
-  if (!p_env->gen.toc_init) 
-    read_toc_win32ioctl (p_env);
 
-  if (!p_env->gen.toc_init) 
-    return CDIO_DISC_MODE_NO_INFO;
+/*! 
+  Get disc type associated with cd object.
+*/
+discmode_t
+get_discmode_win32ioctl (_img_private_t *p_env)
+{
+  track_t i_track;
+  discmode_t discmode;
+
+  if (!p_env) return CDIO_DISC_MODE_ERROR;
+  
+  discmode = dvd_discmode_win32ioctl(p_env);
+
+  if (CDIO_DISC_MODE_NO_INFO != discmode) return discmode;
+  
+  if (!p_env->gen.toc_init) read_toc_win32ioctl (p_env);
+
+  if (!p_env->gen.toc_init) return CDIO_DISC_MODE_ERROR;
 
   for (i_track = p_env->gen.i_first_track; 
        i_track < p_env->gen.i_first_track + p_env->gen.i_tracks ; 
@@ -360,11 +599,11 @@ is_cdrom_win32ioctl(const char c_drive_letter)
    Reads an audio device using the DeviceIoControl method into data
    starting from lsn.  Returns 0 if no error.
  */
-int
+driver_return_code_t
 read_audio_sectors_win32ioctl (_img_private_t *p_env, void *data, lsn_t lsn, 
 			       unsigned int nblocks) 
 {
-  DWORD dwBytesReturned;
+  DWORD dw_bytes_returned;
   RAW_READ_INFO cdrom_raw;
   
   /* Initialize CDROM_RAW_READ structure */
@@ -376,7 +615,7 @@ read_audio_sectors_win32ioctl (_img_private_t *p_env, void *data, lsn_t lsn,
 		       IOCTL_CDROM_RAW_READ, &cdrom_raw,
 		       sizeof(RAW_READ_INFO), data,
 		       CDIO_CD_FRAMESIZE_RAW * nblocks,
-		       &dwBytesReturned, NULL ) == 0 ) {
+		       &dw_bytes_returned, NULL ) == 0 ) {
     char *psz_msg = NULL;
     long int i_err = GetLastError();
     FORMAT_ERROR(i_err, psz_msg);
@@ -388,9 +627,9 @@ read_audio_sectors_win32ioctl (_img_private_t *p_env, void *data, lsn_t lsn,
 		(long unsigned int) lsn, i_err);
     }
     LocalFree(psz_msg);
-    return 1;
+    return DRIVER_OP_ERROR;
   }
-  return 0;
+  return DRIVER_OP_SUCCESS;
 }
 
 /*!
@@ -400,7 +639,7 @@ read_audio_sectors_win32ioctl (_img_private_t *p_env, void *data, lsn_t lsn,
 static int
 read_raw_sector (_img_private_t *p_env, void *p_buf, lsn_t lsn) 
 {
-  scsi_mmc_cdb_t cdb = {{0, }};
+  mmc_cdb_t cdb = {{0, }};
 
   /* ReadCD CDB12 command.  The values were taken from MMC1 draft paper. */
   CDIO_MMC_SET_COMMAND      (cdb.field, CDIO_MMC_GPCMD_READ_CD);
@@ -409,10 +648,10 @@ read_raw_sector (_img_private_t *p_env, void *p_buf, lsn_t lsn)
 
   cdb.field[9]=0xF8;  /* Raw read, 2352 bytes per sector */
   
-  return run_scsi_cmd_win32ioctl(p_env, OP_TIMEOUT_MS,
-				     scsi_mmc_get_cmd_len(cdb.field[0]), 
-				     &cdb, SCSI_MMC_DATA_READ, 
-				     CDIO_CD_FRAMESIZE_RAW, p_buf);
+  return run_mmc_cmd_win32ioctl(p_env, OP_TIMEOUT_MS,
+				mmc_get_cmd_len(cdb.field[0]), 
+				&cdb, SCSI_MMC_DATA_READ, 
+				CDIO_CD_FRAMESIZE_RAW, p_buf);
 }
 
 /*!
@@ -420,7 +659,7 @@ read_raw_sector (_img_private_t *p_env, void *p_buf, lsn_t lsn)
    data starting from lsn. Returns 0 if no error.
  */
 int
-read_mode2_sector_win32ioctl (const _img_private_t *p_env, void *p_data, 
+read_mode2_sector_win32ioctl (_img_private_t *p_env, void *p_data, 
 			      lsn_t lsn, bool b_form2)
 {
   char buf[CDIO_CD_FRAMESIZE_RAW] = { 0, };
@@ -441,7 +680,7 @@ read_mode2_sector_win32ioctl (const _img_private_t *p_env, void *p_data,
    data starting from lsn. Returns 0 if no error.
  */
 int
-read_mode1_sector_win32ioctl (const _img_private_t *env, void *data, 
+read_mode1_sector_win32ioctl (_img_private_t *env, void *data, 
 			      lsn_t lsn, bool b_form2)
 {
   char buf[CDIO_CD_FRAMESIZE_RAW] = { 0, };
@@ -555,7 +794,7 @@ init_win32ioctl (_img_private_t *env)
 static bool
 read_fulltoc_win32mmc (_img_private_t *p_env) 
 {
-  scsi_mmc_cdb_t  cdb = {{0, }};
+  mmc_cdb_t  cdb = {{0, }};
   CDROM_TOC_FULL  cdrom_toc_full;
   int             i_status, i, j;
   int             i_track_format = 0;
@@ -574,10 +813,10 @@ read_fulltoc_win32mmc (_img_private_t *p_env)
   /* Setup to read header, to get length of data */
   CDIO_MMC_SET_READ_LENGTH16(cdb.field, sizeof(cdrom_toc_full));
 
-  i_status = run_scsi_cmd_win32ioctl (p_env, 1000*60*3,
-				scsi_mmc_get_cmd_len(cdb.field[0]), 
-				&cdb, SCSI_MMC_DATA_READ, 
-				sizeof(cdrom_toc_full), &cdrom_toc_full);
+  i_status = run_mmc_cmd_win32ioctl (p_env, 1000*60*3,
+				     mmc_get_cmd_len(cdb.field[0]), 
+				     &cdb, SCSI_MMC_DATA_READ, 
+				     sizeof(cdrom_toc_full), &cdrom_toc_full);
 
   if ( 0 != i_status ) {
     cdio_debug ("SCSI MMC READ_TOC failed\n");  
@@ -605,10 +844,13 @@ read_fulltoc_win32mmc (_img_private_t *p_env)
     if ( 0xA2 ==  j ) { 
       /* Start position of the lead out */
       p_env->tocent[ p_env->gen.i_tracks ].start_lsn = 
-	cdio_msf3_to_lba(
-			 cdrom_toc_full.TrackData[i].PMIN,
-			 cdrom_toc_full.TrackData[i].PSEC,
-			 cdrom_toc_full.TrackData[i].PFRAME );
+	cdio_lba_to_lsn(
+			cdio_msf3_to_lba(
+					 cdrom_toc_full.TrackData[i].PMIN,
+					 cdrom_toc_full.TrackData[i].PSEC,
+					 cdrom_toc_full.TrackData[i].PFRAME 
+					 )
+			);
       p_env->tocent[ p_env->gen.i_tracks ].Control 
 	= cdrom_toc_full.TrackData[i].Control;
       p_env->tocent[ p_env->gen.i_tracks ].Format  = i_track_format;
@@ -618,10 +860,13 @@ read_fulltoc_win32mmc (_img_private_t *p_env)
     if (cdrom_toc_full.TrackData[i].POINT > 0 
 	&& cdrom_toc_full.TrackData[i].POINT <= p_env->gen.i_tracks) {
       p_env->tocent[j-1].start_lsn = 
-	cdio_msf3_to_lba(
-			 cdrom_toc_full.TrackData[i].PMIN,
-			 cdrom_toc_full.TrackData[i].PSEC,
-			 cdrom_toc_full.TrackData[i].PFRAME );
+	cdio_lba_to_lsn(
+			cdio_msf3_to_lba(
+					 cdrom_toc_full.TrackData[i].PMIN,
+					 cdrom_toc_full.TrackData[i].PSEC,
+					 cdrom_toc_full.TrackData[i].PFRAME 
+					 )
+			);
       p_env->tocent[j-1].Control = 
 	cdrom_toc_full.TrackData[i].Control;
       p_env->tocent[j-1].Format  = i_track_format;
@@ -653,27 +898,55 @@ bool
 read_toc_win32ioctl (_img_private_t *p_env) 
 {
   CDROM_TOC    cdrom_toc;
-  DWORD        dwBytesReturned;
+  DWORD        dw_bytes_returned;
   unsigned int i, j;
-
+  bool         b_fulltoc_first;  /* Do we do fulltoc or DeviceIoControl 
+				    first? */
   if ( ! p_env ) return false;
 
-  if ( read_fulltoc_win32mmc(p_env) ) return true;
+  /* 
+     The MMC5 spec says:
+       For media other than CD, information may be fabricated in order
+                                            ^^^ ^^
+       to emulate a CD structure for the specific media.
+
+     There is no requirement though that it *has* to and some DVD
+     drives like one by Thompson for XBOX don't support a
+     IOCTL_CDROM_READ_TOC for DVD's. So if we have a DVD we should not
+     prefer getting the TOC via MMC.
+
+     But on the other hand in GNU/Linux it is reported that using the
+     TOC via MMC gives better information such as for CD DATA Form 2 (used
+     in SVCDs). So if we *don't* have a DVD I think we want to try MMC
+     first. 
+
+     Is this complicated enough? I could be wrong...
+
+   */
+  b_fulltoc_first = (CDIO_DISC_MODE_NO_INFO == dvd_discmode_win32ioctl(p_env));
+
+  if ( b_fulltoc_first && read_fulltoc_win32mmc(p_env) ) return true;
+
+  /* SCSI-MMC READ_TOC (FULTOC) read failed or we don't want to try it
+     initiaily.  Try reading TOC via DeviceIoControl... */
 
-  /* SCSI-MMC READ_TOC (FULTOC) read failed.  Try reading TOC via
-     DeviceIoControl instead */
   if( DeviceIoControl( p_env->h_device_handle,
 		       IOCTL_CDROM_READ_TOC,
 		       NULL, 0, &cdrom_toc, sizeof(CDROM_TOC),
-		       &dwBytesReturned, NULL ) == 0 ) {
+		       &dw_bytes_returned, NULL ) == 0 ) {
     char *psz_msg = NULL;
     long int i_err = GetLastError();
+    cdio_log_level_t loglevel = b_fulltoc_first 
+      ? CDIO_LOG_WARN : CDIO_LOG_DEBUG;
+
     FORMAT_ERROR(i_err, psz_msg);
     if (psz_msg) {
-      cdio_warn("could not read TOC (%ld): %s", i_err, psz_msg);
+      cdio_log(loglevel, "could not read TOC (%ld): %s", i_err, psz_msg);
       LocalFree(psz_msg);
     } else 
-      cdio_warn("could not read TOC (%ld)", i_err);
+      cdio_log(loglevel, "could not read TOC (%ld)", i_err);
+
+    if ( !b_fulltoc_first && read_fulltoc_win32mmc(p_env) ) return true;
     return false;
   }
   
@@ -683,11 +956,13 @@ read_toc_win32ioctl (_img_private_t *p_env)
   j = p_env->gen.i_first_track;
   for( i = 0 ; i <= p_env->gen.i_tracks ; i++, j++ ) {
     p_env->tocent[ i ].start_lsn = 
-      cdio_msf3_to_lba( cdrom_toc.TrackData[i].Address[1],
-			cdrom_toc.TrackData[i].Address[2],
-			cdrom_toc.TrackData[i].Address[3] );
+      cdio_lba_to_lsn(
+		      cdio_msf3_to_lba( cdrom_toc.TrackData[i].Address[1],
+					cdrom_toc.TrackData[i].Address[2],
+					cdrom_toc.TrackData[i].Address[3] )
+		      );
     p_env->tocent[i].Control   = cdrom_toc.TrackData[i].Control;
-    p_env->tocent[i].Format    = cdrom_toc.TrackData[i].Format;
+    p_env->tocent[i].Format    = cdrom_toc.TrackData[i].Adr;
 
     p_env->gen.track_flags[j].preemphasis = 
       p_env->tocent[i].Control & 0x1 
@@ -716,9 +991,9 @@ read_toc_win32ioctl (_img_private_t *p_env)
 
  */
 char *
-get_mcn_win32ioctl (const _img_private_t *env) {
+get_mcn_win32ioctl (const _img_private_t *p_env) {
 
-  DWORD dwBytesReturned;
+  DWORD dw_bytes_returned;
   SUB_Q_MEDIA_CATALOG_NUMBER mcn;
   CDROM_SUB_Q_DATA_FORMAT q_data_format;
   
@@ -726,13 +1001,16 @@ get_mcn_win32ioctl (const _img_private_t *env) {
   
   q_data_format.Format = CDIO_SUBCHANNEL_MEDIA_CATALOG;
 
-  q_data_format.Track=1;
+  /* MSDN info on CDROM_SUB_Q_DATA_FORMAT says if Format is set to 
+     get MCN, track must be set 0.
+   */
+  q_data_format.Track=0; 
   
-  if( DeviceIoControl( env->h_device_handle,
+  if( ! DeviceIoControl( p_env->h_device_handle,
 		       IOCTL_CDROM_READ_Q_CHANNEL,
 		       &q_data_format, sizeof(q_data_format), 
 		       &mcn, sizeof(mcn),
-		       &dwBytesReturned, NULL ) == 0 ) {
+		       &dw_bytes_returned, NULL ) ) {
     cdio_warn( "could not read Q Channel at track %d", 1);
   } else if (mcn.Mcval) 
     return strdup(mcn.MediaCatalog);
