@@ -1,8 +1,8 @@
 /*
-    $Id: iso9660.c,v 1.1 2004/12/18 17:29:32 rocky Exp $
+    $Id: iso9660.c,v 1.13 2005/03/03 10:32:01 rocky Exp $
 
     Copyright (C) 2000 Herbert Valerio Riedel <hvr@gnu.org>
-    Copyright (C) 2003, 2004 Rocky Bernstein <rocky@panix.com>
+    Copyright (C) 2003, 2004, 2005 Rocky Bernstein <rocky@panix.com>
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -19,6 +19,13 @@
     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
 
+/*! String inside frame which identifies an ISO 9660 filesystem. This
+    string is the "id" field of an iso9660_pvd_t or an iso9660_svd_t.
+    Note should come *before* #include <cdio/iso9660.h> which does 
+    a #define of this name.
+*/
+const char ISO_STANDARD_ID[] = {'C', 'D', '0', '0', '1'};
+
 /* Private headers */
 #include "iso9660_private.h"
 #include "cdio_assert.h"
@@ -37,7 +44,17 @@
 #include <stdio.h>
 #endif
 
-static const char _rcsid[] = "$Id: iso9660.c,v 1.1 2004/12/18 17:29:32 rocky Exp $";
+#ifdef HAVE_ERRNO_H
+#include <errno.h>
+#endif
+
+static const char _rcsid[] = "$Id: iso9660.c,v 1.13 2005/03/03 10:32:01 rocky Exp $";
+
+/* Variables to hold debugger-helping enumerations */
+enum iso_enum1_s     iso_enums1;
+enum iso_flag_enum_s iso_flag_enums;
+enum iso_vd_enum_s   iso_vd_enums;
+enum iso_extension_enum_s iso_extension_enums;
 
 /* some parameters... */
 #define SYSTEM_ID         "CD-RTOS CD-BRIDGE"
@@ -81,47 +98,125 @@ pathtable_get_size_and_entries(const void *pt, unsigned int *size,
   If tm is to reflect the localtime set b_localtime true, otherwise
   tm will reported in GMT.
 */
-void
+bool
 iso9660_get_dtime (const iso9660_dtime_t *idr_date, bool b_localtime,
                    /*out*/ struct tm *p_tm)
 {
-  time_t t;
-  struct tm *p_temp_tm;
-  
-  if (!idr_date) return;
+  if (!idr_date) return false;
 
+  /* 
+     Section 9.1.5 of ECMA 119 says: 
+     If all seven numbers are zero, it shall mean that the date and
+     time are not specified. 
+
+     HACK: However we've seen it happen that everything except gmtoff
+     is zero and the expected date is the beginning of the epoch. So
+     we accept 6 numbers being zero. I'm also not sure if using the
+     beginning of the Epoch is also the right thing to do either.
+  */
+
+  if ( 0 == idr_date->dt_year   && 0 == idr_date->dt_month &&
+       0 == idr_date->dt_day    && 0 == idr_date->dt_hour  &&
+       0 == idr_date->dt_minute && 0 == idr_date->dt_second ) {
+    time_t t = 0;
+    struct tm *p_temp_tm;
+    p_temp_tm = localtime(&t);
+    
+    memcpy(p_tm, p_temp_tm, sizeof(struct tm));
+    return true;
+  }
+  
   memset(p_tm, 0, sizeof(struct tm));
+
   p_tm->tm_year   = idr_date->dt_year;
   p_tm->tm_mon    = idr_date->dt_month - 1;
   p_tm->tm_mday   = idr_date->dt_day;
   p_tm->tm_hour   = idr_date->dt_hour;
   p_tm->tm_min    = idr_date->dt_minute;
   p_tm->tm_sec    = idr_date->dt_second;
+  p_tm->tm_isdst  = -1; /* information not available */
+  {
+    time_t t = 0;
+    struct tm *p_temp_tm = b_localtime ? localtime(&t): gmtime(&t);
+    
+    /* Recompute tm_wday and tm_yday via mktime. Also adjust for 
+       time-zone differences */
+    t = mktime(p_tm);
+    t -= idr_date->dt_gmtoff * (15 * 60);
+#ifdef HAVE_TM_GMTOFF
+    t += p_temp_tm->tm_gmtoff;
+#else 
+    {
+      static bool i_first_time = true;
+      static time_t gmt_off;
+      if (i_first_time) {
+        struct tm tm_local, tm_gmt;
+        memcpy(&tm_local, localtime(&t), sizeof(struct tm));
+        memcpy(&tm_gmt  , gmtime(&t)   , sizeof(struct tm));
+        gmt_off      = mktime(&tm_local) - mktime(&tm_gmt);
+        i_first_time = false;
+      }
+      t += gmt_off;
+    }
+#endif
+    
+    p_temp_tm = b_localtime ? localtime(&t) : gmtime(&t);
 
-#if defined(HAVE_TM_GMTOFF) && defined(HAVE_TZSET)
-  if (b_localtime) {
-    tzset();
-#if defined(HAVE_TZNAME)
-    p_tm->tm_zone   = (char *) tzname;
-#endif
-#if defined(HAVE_DAYLIGHT)
-    p_tm->tm_isdst  = daylight;
-    p_tm->tm_gmtoff = timezone;
-#endif
+    if (p_temp_tm->tm_isdst) {
+      t += 60*60;
+      p_temp_tm = b_localtime ? localtime(&t) : gmtime(&t);
+    }
+    
+    memcpy(p_tm, p_temp_tm, sizeof(struct tm));
   }
+
+
+  return true;
+}
+
+#define set_ltime_field(TM_FIELD, LT_FIELD, ADD_CONSTANT)               \
+  {                                                                     \
+    p_tm->TM_FIELD = strtol(p_ldate->LT_FIELD,                          \
+                            (char **)NULL, 10)+ADD_CONSTANT;            \
+    if (0 != errno) return false;                                       \
+  }
+    
+/*!
+  Get "long" time in format used in ISO 9660 primary volume descriptor
+  from a Unix time structure. 
+*/
+bool
+iso9660_get_ltime (const iso9660_ltime_t *p_ldate, 
+                   /*out*/ struct tm *p_tm)
+{
+  if (!p_tm) return false;
+  memset(p_tm, 0, sizeof(struct tm));
+  set_ltime_field(tm_year, lt_year,   0);
+  set_ltime_field(tm_mon,  lt_month, -1);
+  set_ltime_field(tm_mday, lt_day,    0);
+  set_ltime_field(tm_hour, lt_hour,   0);
+  set_ltime_field(tm_min,  lt_minute, 0);
+  set_ltime_field(tm_sec,  lt_second, 0);
+  p_tm->tm_isdst= -1; /* information not available */
+#ifdef HAVE_TM_GMTOFF
+  p_tm->tm_gmtoff = - p_ldate->lt_gmtoff * (15 * 60);
 #endif
 
   /* Recompute tm_wday and tm_yday via mktime. */
-  t = mktime(p_tm);
-
-  if (b_localtime)
-    p_temp_tm = localtime(&t);
-  else
+  {
+    time_t t;
+    struct tm *p_temp_tm;
+    
+    t = mktime(p_tm);
     p_temp_tm = gmtime(&t);
-
-  memcpy(p_tm, p_temp_tm, sizeof(struct tm));
+    
+    p_tm->tm_wday = p_temp_tm->tm_wday;
+    p_tm->tm_yday = p_temp_tm->tm_yday;
+  }
+  return true;
 }
 
+
 /*!
   Set time in format used in ISO 9660 directory index record
   from a Unix time structure. */
@@ -210,7 +305,8 @@ iso9660_name_translate_ext(const char *old, char *new, uint8_t i_joliet_level)
 {
   int len = strlen(old);
   int i;
-  
+
+  if (0 == len) return 0;
   for (i = 0; i < len; i++) {
     unsigned char c = old[i];
     if (!c)
@@ -348,9 +444,9 @@ iso9660_isachar (int c)
 void 
 iso9660_set_evd(void *pd)
 {
-  struct iso_volume_descriptor ied;
+  iso_volume_descriptor_t ied;
 
-  cdio_assert (sizeof(struct iso_volume_descriptor) == ISO_BLOCKSIZE);
+  cdio_assert (sizeof(iso_volume_descriptor_t) == ISO_BLOCKSIZE);
 
   cdio_assert (pd != NULL);
   
@@ -578,14 +674,14 @@ iso9660_dir_init_new_su (void *dir,
 void 
 iso9660_pathtable_init (void *pt)
 {
-  cdio_assert (sizeof (struct iso_path_table) == 8);
+  cdio_assert (sizeof (iso_path_table_t) == 8);
 
   cdio_assert (pt != NULL);
   
   memset (pt, 0, ISO_BLOCKSIZE); /* fixme */
 }
 
-static const struct iso_path_table*
+static const iso_path_table_t *
 pathtable_get_entry (const void *pt, unsigned int entrynum)
 {
   const uint8_t *tmp = pt;
@@ -601,7 +697,7 @@ pathtable_get_entry (const void *pt, unsigned int entrynum)
 
       cdio_assert (count < entrynum);
 
-      offset += sizeof (struct iso_path_table);
+      offset += sizeof (iso_path_table_t);
       offset += from_711 (*tmp);
       if (offset % 2)
         offset++;
@@ -628,7 +724,7 @@ pathtable_get_size_and_entries (const void *pt,
 
   while (from_711 (*tmp)) 
     {
-      offset += sizeof (struct iso_path_table);
+      offset += sizeof (iso_path_table_t);
       offset += from_711 (*tmp);
       if (offset % 2)
         offset++;
@@ -657,14 +753,14 @@ iso9660_pathtable_l_add_entry (void *pt,
                                uint32_t extent, 
                                uint16_t parent)
 {
-  struct iso_path_table *ipt = 
-    (struct iso_path_table*)((char *)pt + iso9660_pathtable_get_size (pt));
+  iso_path_table_t *ipt = 
+    (iso_path_table_t *)((char *)pt + iso9660_pathtable_get_size (pt));
   size_t name_len = strlen (name) ? strlen (name) : 1;
   unsigned int entrynum = 0;
 
   cdio_assert (iso9660_pathtable_get_size (pt) < ISO_BLOCKSIZE); /*fixme */
 
-  memset (ipt, 0, sizeof (struct iso_path_table) + name_len); /* paranoia */
+  memset (ipt, 0, sizeof (iso_path_table_t) + name_len); /* paranoia */
 
   ipt->name_len = to_711 (name_len);
   ipt->extent = to_731 (extent);
@@ -675,7 +771,7 @@ iso9660_pathtable_l_add_entry (void *pt,
 
   if (entrynum > 1)
     {
-      const struct iso_path_table *ipt2 
+      const iso_path_table_t *ipt2 
         = pathtable_get_entry (pt, entrynum - 2);
 
       cdio_assert (ipt2 != NULL);
@@ -692,14 +788,14 @@ iso9660_pathtable_m_add_entry (void *pt,
                                uint32_t extent, 
                                uint16_t parent)
 {
-  struct iso_path_table *ipt =
-    (struct iso_path_table*)((char *)pt + iso9660_pathtable_get_size (pt));
+  iso_path_table_t *ipt =
+    (iso_path_table_t *)((char *)pt + iso9660_pathtable_get_size (pt));
   size_t name_len = strlen (name) ? strlen (name) : 1;
   unsigned int entrynum = 0;
 
   cdio_assert (iso9660_pathtable_get_size(pt) < ISO_BLOCKSIZE); /* fixme */
 
-  memset(ipt, 0, sizeof (struct iso_path_table) + name_len); /* paranoia */
+  memset(ipt, 0, sizeof (iso_path_table_t) + name_len); /* paranoia */
 
   ipt->name_len = to_711 (name_len);
   ipt->extent = to_732 (extent);
@@ -710,7 +806,7 @@ iso9660_pathtable_m_add_entry (void *pt,
 
   if (entrynum > 1)
     {
-      const struct iso_path_table *ipt2 
+      const iso_path_table_t *ipt2 
         = pathtable_get_entry (pt, entrynum - 2);
 
       cdio_assert (ipt2 != NULL);
