@@ -1,8 +1,8 @@
 /*
-    $Id: iso9660_fs.c,v 1.4 2005/01/20 01:00:52 rocky Exp $
+    $Id: iso9660_fs.c,v 1.28 2005/04/15 05:06:09 rocky Exp $
 
     Copyright (C) 2001 Herbert Valerio Riedel <hvr@gnu.org>
-    Copyright (C) 2003, 2004 Rocky Bernstein <rocky@panix.com>
+    Copyright (C) 2003, 2004, 2005 Rocky Bernstein <rocky@panix.com>
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -18,6 +18,7 @@
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
+/* iso9660 filesystem-based routines */
 
 #ifdef HAVE_CONFIG_H
 # include "config.h"
@@ -51,20 +52,153 @@
 
 #include <stdio.h>
 
-static const char _rcsid[] = "$Id: iso9660_fs.c,v 1.4 2005/01/20 01:00:52 rocky Exp $";
+static const char _rcsid[] = "$Id: iso9660_fs.c,v 1.28 2005/04/15 05:06:09 rocky Exp $";
 
 /* Implementation of iso9660_t type */
-struct _iso9660 {
+struct _iso9660_s {
   CdioDataSource_t *stream; /* Stream pointer */
-  bool b_xa;              /* true if has XA attributes. */
-  uint8_t  i_joliet_level;/* 0 = no Joliet extensions.
-			     1-3: Joliet level. */
+  bool_3way_t b_xa;         /* true if has XA attributes. */
+  bool_3way_t b_mode2;      /* true if has mode 2, false for mode 1. */
+  uint8_t  i_joliet_level;  /* 0 = no Joliet extensions.
+			       1-3: Joliet level. */
   iso9660_pvd_t pvd;      
   iso9660_svd_t svd;      
   iso_extension_mask_t iso_extension_mask; /* What extensions we
 					      tolerate. */
+  uint32_t i_datastart;     /* Usually 0 when i_framesize is ISO_BLOCKSIZE.
+			       This is the normal condition. But in a fuzzy 
+			       read we may be reading a CD-image
+			       and not a true ISO 9660 image this might be
+			       CDIO_CD_SYNC_SIZE
+                            */ 
+  uint32_t i_framesize;     /* Usually ISO_BLOCKSIZE (2048), but in a
+			      fuzzy read, we may be reading a CD-image
+			      and not a true ISO 9660 image this might
+			      be CDIO_CD_FRAMESIZE_RAW (2352) or
+			      M2RAW_SECTOR_SIZE (2336).
+                            */ 
+  int i_fuzzy_offset;       /* Adjustment in bytes to make ISO_STANDARD_ID
+			       ("CD001") come out as ISO_PVD_SECTOR
+			       (frame 16).  Normally this should be 0
+			       for an ISO 9660 image, but if one is
+			       say reading a BIN/CUE or cdrdao BIN/TOC
+			       without having the CUE or TOC and
+			       trying to extract an ISO-9660
+			       filesystem inside that it may be
+			       different.
+			     */
 };
 
+static long int iso9660_seek_read_framesize (const iso9660_t *p_iso, 
+					     void *ptr, lsn_t start, 
+					     long int size, 
+					     uint16_t i_framesize);
+
+/* Adjust the p_iso's i_datastart, i_byte_offset and i_framesize 
+   based on whether we find a frame header or not.
+*/
+static void 
+adjust_fuzzy_pvd( iso9660_t *p_iso )
+{
+  long int i_byte_offset;
+  
+  if (!p_iso) return;
+
+  i_byte_offset = (ISO_PVD_SECTOR * p_iso->i_framesize) 
+    + p_iso->i_fuzzy_offset + p_iso->i_datastart;
+
+  /* If we have a raw 2352-byte frame then we should expect to see a sync
+     frame and a header. 
+   */
+  if (CDIO_CD_FRAMESIZE_RAW == p_iso->i_framesize) {
+    const int pre_user_data=CDIO_CD_SYNC_SIZE +  CDIO_CD_HEADER_SIZE 
+      + CDIO_CD_SUBHEADER_SIZE;
+    char buf[pre_user_data];
+    
+    i_byte_offset -= pre_user_data;
+    
+    if ( DRIVER_OP_SUCCESS != cdio_stream_seek (p_iso->stream, i_byte_offset, 
+						SEEK_SET) )
+      return;
+    if (sizeof(buf) == cdio_stream_read (p_iso->stream, buf, sizeof(buf), 1)) {
+      /* Does the sector frame header suggest Mode 1 format? */
+      if (!memcmp(CDIO_SECTOR_SYNC_HEADER, buf+CDIO_CD_SUBHEADER_SIZE, 
+		  CDIO_CD_SYNC_SIZE)) {
+	if (buf[14+CDIO_CD_SUBHEADER_SIZE] != 0x16) {
+	  cdio_warn ("Expecting the PVD sector header MSF to be 0x16, is: %x", 
+		     buf[14]);
+	}
+	if (buf[15+CDIO_CD_SUBHEADER_SIZE] != 0x1) {
+	  cdio_warn ("Expecting the PVD sector mode to be Mode 1 is: %x", 
+		     buf[15]);
+	}
+	p_iso->b_mode2 = nope;
+	p_iso->b_xa = nope;
+      } else if (!memcmp(CDIO_SECTOR_SYNC_HEADER, buf, CDIO_CD_SYNC_SIZE)) {
+	/* Frame header indicates Mode 2 Form 1*/
+	if (buf[14] != 0x16) {
+	  cdio_warn ("Expecting the PVD sector header MSF to be 0x16, is: %x", 
+		     buf[14]);
+	}
+	if (buf[15] != 0x2) {
+	  cdio_warn ("Expecting the PVD sector mode to be Mode 2 is: %x", 
+		     buf[15]);
+	}
+	p_iso->b_mode2 = yep;
+	/* Do do: check Mode 2 Form 2? */
+      } else {
+	  /* Has no frame header */
+	  p_iso->i_framesize = M2RAW_SECTOR_SIZE;
+	  p_iso->i_fuzzy_offset = (CDIO_CD_FRAMESIZE_RAW - M2RAW_SECTOR_SIZE) 
+	    * ISO_PVD_SECTOR + p_iso->i_fuzzy_offset + p_iso->i_datastart;
+	  p_iso->i_datastart = 0;
+	}
+    }
+  }
+
+  
+}
+
+/*!
+  Open an ISO 9660 image for reading in either fuzzy mode or not.
+*/
+static iso9660_t *
+iso9660_open_ext_private (const char *pathname,
+			  iso_extension_mask_t iso_extension_mask,
+			  uint16_t i_fuzz, bool b_fuzzy)
+{
+  iso9660_t *p_iso = (iso9660_t *) calloc(1, sizeof(iso9660_t)) ;
+  bool b_have_superblock;
+
+  if (!p_iso) return NULL;
+  
+  p_iso->stream = cdio_stdio_new( pathname );
+  if (NULL == p_iso->stream) 
+    goto error;
+
+  p_iso->i_framesize = ISO_BLOCKSIZE;
+
+  b_have_superblock = (b_fuzzy) 
+    ? iso9660_ifs_fuzzy_read_superblock(p_iso, iso_extension_mask, i_fuzz)
+    : iso9660_ifs_read_superblock(p_iso, iso_extension_mask) ;
+
+  if ( ! b_have_superblock ) goto error;
+  
+  /* Determine if image has XA attributes. */
+  
+  p_iso->b_xa = strncmp ((char *) &(p_iso->pvd) + ISO_XA_MARKER_OFFSET, 
+			 ISO_XA_MARKER_STRING,
+			 strlen (ISO_XA_MARKER_STRING)) 
+    ? nope : yep;
+  
+  p_iso->iso_extension_mask = iso_extension_mask;
+  return p_iso;
+
+ error:
+  free(p_iso);
+  return NULL;
+}
+
 /*!
   Open an ISO 9660 image for reading. Maybe in the future we will have
   a mode. NULL is returned on error.
@@ -80,34 +214,35 @@ iso9660_open (const char *pathname /*, mode*/)
   a mode. NULL is returned on error.
 */
 iso9660_t *
-iso9660_open_ext (const char *pathname,
+iso9660_open_ext (const char *psz_path,
 		  iso_extension_mask_t iso_extension_mask)
 {
-  iso9660_t *p_iso = (iso9660_t *) _cdio_malloc(sizeof(struct _iso9660)) ;
-
-  if (NULL == p_iso) return NULL;
-  
-  p_iso->stream = cdio_stdio_new( pathname );
-  if (NULL == p_iso->stream) 
-    goto error;
-  
-  if ( !iso9660_ifs_read_superblock(p_iso, iso_extension_mask) )
-    goto error;
-  
-  /* Determine if image has XA attributes. */
-  
-  p_iso->b_xa = !strncmp ((char *) &(p_iso->pvd) + ISO_XA_MARKER_OFFSET, 
-			  ISO_XA_MARKER_STRING, 
-			  strlen (ISO_XA_MARKER_STRING));
-  p_iso->iso_extension_mask = iso_extension_mask;
-  return p_iso;
-
- error:
-  free(p_iso);
-  return NULL;
+  return iso9660_open_ext_private(psz_path, iso_extension_mask, 0, false);
 }
 
 
+/*!
+  Open an ISO 9660 image for reading. Maybe in the future we will have
+  a mode. NULL is returned on error.
+*/
+iso9660_t *
+iso9660_open_fuzzy (const char *psz_path, uint16_t i_fuzz /*, mode*/)
+{
+  return iso9660_open_fuzzy_ext(psz_path, ISO_EXTENSION_NONE, i_fuzz);
+}
+
+/*!
+  Open an ISO 9660 image for reading. Maybe in the future we will have
+  a mode. NULL is returned on error.
+*/
+iso9660_t *
+iso9660_open_fuzzy_ext (const char *psz_path,
+			iso_extension_mask_t iso_extension_mask,
+			uint16_t i_fuzz)
+{
+  return iso9660_open_ext_private(psz_path, iso_extension_mask, i_fuzz,
+				  true);
+}
 
 /*!
   Close previously opened ISO 9660 image.
@@ -125,16 +260,16 @@ iso9660_close (iso9660_t *p_iso)
 }
 
 static bool
-check_pvd (const iso9660_pvd_t *p_pvd) 
+check_pvd (const iso9660_pvd_t *p_pvd, cdio_log_level_t log_level) 
 {
   if ( ISO_VD_PRIMARY != from_711(p_pvd->type) ) {
-    cdio_warn ("unexpected PVD type %d", p_pvd->type);
+    cdio_log (log_level, "unexpected PVD type %d", p_pvd->type);
     return false;
   }
   
   if (strncmp (p_pvd->id, ISO_STANDARD_ID, strlen (ISO_STANDARD_ID)))
     {
-      cdio_warn ("unexpected ID encountered (expected `"
+      cdio_log (log_level, "unexpected ID encountered (expected `"
 		ISO_STANDARD_ID "', got `%.5s'", p_pvd->id);
       return false;
     }
@@ -381,11 +516,27 @@ bool iso9660_ifs_get_volumeset_id(iso9660_t *p_iso,
       return true;
   }
 #endif /*HAVE_JOLIET*/
-  *p_psz_volumeset_id = iso9660_get_volume_id( &(p_iso->pvd) );
+  *p_psz_volumeset_id = iso9660_get_volumeset_id( &(p_iso->pvd) );
   return *p_psz_volumeset_id != NULL && strlen(*p_psz_volumeset_id);
 }
 
 
+/*!
+  Read the Primary Volume Descriptor for an ISO 9660 image.
+  True is returned if read, and false if there was an error.
+*/
+static bool 
+iso9660_ifs_read_pvd_loglevel (const iso9660_t *p_iso, 
+			       /*out*/ iso9660_pvd_t *p_pvd, 
+			       cdio_log_level_t log_level)
+{
+  if (0 == iso9660_iso_seek_read (p_iso, p_pvd, ISO_PVD_SECTOR, 1)) {
+    cdio_log ( log_level, "error reading PVD sector (%d)", ISO_PVD_SECTOR );
+    return false;
+  }
+  return check_pvd(p_pvd, log_level);
+}
+
 /*!
   Read the Primary Volume Descriptor for an ISO 9660 image.
   True is returned if read, and false if there was an error.
@@ -393,11 +544,7 @@ bool iso9660_ifs_get_volumeset_id(iso9660_t *p_iso,
 bool 
 iso9660_ifs_read_pvd (const iso9660_t *p_iso, /*out*/ iso9660_pvd_t *p_pvd)
 {
-  if (0 == iso9660_iso_seek_read (p_iso, p_pvd, ISO_PVD_SECTOR, 1)) {
-    cdio_warn ("error reading PVD sector (%d)", ISO_PVD_SECTOR);
-    return false;
-  }
-  return check_pvd(p_pvd);
+  return iso9660_ifs_read_pvd_loglevel(p_iso, p_pvd, CDIO_LOG_WARN);
 }
 
 
@@ -448,7 +595,74 @@ iso9660_ifs_read_superblock (iso9660_t *p_iso,
   return true;
 }
 
+/*!
+  Read the Super block of an ISO 9660 image but determine framesize
+  and datastart and a possible additional offset. Generally here we are
+  not reading an ISO 9660 image but a CD-Image which contains an ISO 9660
+  filesystem.
+*/
+bool 
+iso9660_ifs_fuzzy_read_superblock (iso9660_t *p_iso, 
+				   iso_extension_mask_t iso_extension_mask,
+				   uint16_t i_fuzz)
+{
+  /* Got some work to do to find ISO_STANDARD_ID ("CD001") */
+  unsigned int i;
+  
+  for (i=0; i<i_fuzz; i++) {
+    unsigned int j;
+    char *pvd = NULL;
 
+    for (j = 0; j <= 1; j++ ) {
+      lsn_t lsn;
+      uint16_t k;
+      const uint16_t framesizes[] = { ISO_BLOCKSIZE, CDIO_CD_FRAMESIZE_RAW, 
+				      M2RAW_SECTOR_SIZE } ;
+
+      /* We don't need to loop over a zero offset twice*/
+      if (0==i && j)
+	continue;
+      
+      lsn = (j) ? ISO_PVD_SECTOR - i : ISO_PVD_SECTOR + i;
+      
+      for (k=0; k < 3; k++) {
+	char *p, *q;
+	char frame[CDIO_CD_FRAMESIZE_RAW] = {'\0',};
+	p_iso->i_framesize = framesizes[k];
+	p_iso->i_datastart = (ISO_BLOCKSIZE == framesizes[k]) ? 
+			      0 : CDIO_CD_SYNC_SIZE;
+	p_iso->i_fuzzy_offset = 0;
+	if (0 == iso9660_seek_read_framesize (p_iso, frame, lsn, 1, 
+					      p_iso->i_framesize)) {
+	  return false;
+	}
+
+	q = memchr(frame, 'C', p_iso->i_framesize);
+	for ( p=q; p && p < frame + p_iso->i_framesize ; p=q+1 ) {
+	  q = memchr(p, 'C', p_iso->i_framesize - (p - frame));
+	  if ( !q || (pvd = strstr(q, ISO_STANDARD_ID)) )
+	    break;
+	}
+	
+	if (pvd) {
+	  /* Yay! Found something */
+	  p_iso->i_fuzzy_offset = (pvd - frame - 1) - 
+	    ((ISO_PVD_SECTOR-lsn)*p_iso->i_framesize) ;
+	  /* But is it *really* a PVD? */
+	  if ( iso9660_ifs_read_pvd_loglevel(p_iso, &(p_iso->pvd), 
+					     CDIO_LOG_DEBUG) ) {
+	    adjust_fuzzy_pvd(p_iso);
+	    return true;
+	  }
+	  
+	}
+      }
+    }
+  }
+  return false;
+}
+
+  
 /*!
   Read the Primary Volume Descriptor for of CD.
 */
@@ -456,31 +670,13 @@ bool
 iso9660_fs_read_pvd(const CdIo_t *p_cdio, /*out*/ iso9660_pvd_t *p_pvd)
 {
   /* A bit of a hack, we'll assume track 1 contains ISO_PVD_SECTOR.*/
-  bool b_mode2;
   char buf[CDIO_CD_FRAMESIZE_RAW] = { 0, };
-  int i_rc;
+  driver_return_code_t driver_return = 
+    cdio_read_data_sectors (p_cdio, buf, ISO_PVD_SECTOR, ISO_BLOCKSIZE, 1);
 
-  switch(cdio_get_track_format(p_cdio, 1)) {
-  case TRACK_FORMAT_CDI:
-  case TRACK_FORMAT_XA:
-    b_mode2 = true;
-    break;
-  case TRACK_FORMAT_DATA:
-    b_mode2 = false;
-    break;
-  case TRACK_FORMAT_AUDIO: 
-  case TRACK_FORMAT_PSX: 
-  case TRACK_FORMAT_ERROR: 
-  default:
-    return false;
-  }
-  
-  i_rc = b_mode2 
-      ? cdio_read_mode2_sector (p_cdio, buf, ISO_PVD_SECTOR, false)
-      : cdio_read_mode1_sector (p_cdio, buf, ISO_PVD_SECTOR, false);
-  
-  if (i_rc) {
-    cdio_warn ("error reading PVD sector (%d)", ISO_PVD_SECTOR);
+  if (DRIVER_OP_SUCCESS != driver_return) {
+    cdio_warn ("error reading PVD sector (%d) error %d", ISO_PVD_SECTOR,
+	       driver_return);
     return false;
   }
 
@@ -491,7 +687,7 @@ iso9660_fs_read_pvd(const CdIo_t *p_cdio, /*out*/ iso9660_pvd_t *p_pvd)
   cdio_assert (sizeof(buf) >= sizeof (iso9660_pvd_t));
   memcpy(p_pvd, buf, sizeof(iso9660_pvd_t));
   
-  return check_pvd(p_pvd);
+  return check_pvd(p_pvd, CDIO_LOG_WARN);
 }
 
 
@@ -511,35 +707,18 @@ iso9660_fs_read_superblock (CdIo_t *p_cdio,
     iso9660_pvd_t         *p_pvd = &(p_env->pvd);
     iso9660_svd_t         *p_svd = &(p_env->svd);
     char buf[CDIO_CD_FRAMESIZE_RAW] = { 0, };
-    bool                   b_mode2;
-    int i_rc;
+    driver_return_code_t driver_return;
 
-    /* A bit of a hack, we'll assume track 1 contains ISO_PVD_SECTOR.*/
-    switch(cdio_get_track_format(p_cdio, 1)) {
-    case TRACK_FORMAT_CDI:
-    case TRACK_FORMAT_XA:
-      b_mode2 = true;
-      break;
-    case TRACK_FORMAT_DATA:
-      b_mode2 = false;
-      break;
-    case TRACK_FORMAT_AUDIO: 
-    case TRACK_FORMAT_PSX: 
-    case TRACK_FORMAT_ERROR: 
-    default:
-      return false;
-    }
-  
     if ( !iso9660_fs_read_pvd(p_cdio, p_pvd) )
       return false;
     
     p_env->i_joliet_level = 0;
     
-    i_rc = (b_mode2)
-      ? cdio_read_mode2_sector (p_cdio, buf, ISO_PVD_SECTOR+1, false)
-      : cdio_read_mode1_sector (p_cdio, buf, ISO_PVD_SECTOR+1, false);
+    driver_return = 
+      cdio_read_data_sectors ( p_cdio, buf, ISO_PVD_SECTOR+1, ISO_BLOCKSIZE,
+			       1 );
 
-    if (0 == i_rc) {
+  if (DRIVER_OP_SUCCESS == driver_return) {
       /* The size of a PVD or SVD is smaller than a sector. So we
 	 allocated a bigger block above (buf) and now we'll copy just
 	 the part we need to save.
@@ -578,6 +757,26 @@ iso9660_fs_read_superblock (CdIo_t *p_cdio,
   return true;
 }
 
+/*!
+  Seek to a position and then read n blocks. Size read is returned.
+*/
+static long int 
+iso9660_seek_read_framesize (const iso9660_t *p_iso, void *ptr, 
+			     lsn_t start, long int size, 
+			     uint16_t i_framesize)
+{
+  long int ret;
+  long int i_byte_offset;
+  
+  if (!p_iso) return 0;
+  i_byte_offset = (start * p_iso->i_framesize) + p_iso->i_fuzzy_offset 
+    + p_iso->i_datastart;
+
+  ret = cdio_stream_seek (p_iso->stream, i_byte_offset, SEEK_SET);
+  if (ret!=0) return 0;
+  return cdio_stream_read (p_iso->stream, ptr, i_framesize, size);
+}
+
 /*!
   Seek to a position and then read n blocks. Size read is returned.
 */
@@ -585,92 +784,119 @@ long int
 iso9660_iso_seek_read (const iso9660_t *p_iso, void *ptr, lsn_t start, 
 		       long int size)
 {
-  long int ret;
-  if (NULL == p_iso) return 0;
-  ret = cdio_stream_seek (p_iso->stream, start * ISO_BLOCKSIZE, SEEK_SET);
-  if (ret!=0) return 0;
-  return cdio_stream_read (p_iso->stream, ptr, ISO_BLOCKSIZE, size);
+  return iso9660_seek_read_framesize(p_iso, ptr, start, size, ISO_BLOCKSIZE);
 }
 
 
+
 static iso9660_stat_t *
-_iso9660_dir_to_statbuf (iso9660_dir_t *p_iso9660_dir, 
-			 bool b_mode2, uint8_t i_joliet_level)
+_iso9660_dir_to_statbuf (iso9660_dir_t *p_iso9660_dir, bool_3way_t b_xa, 
+			 uint8_t i_joliet_level)
 {
-  iso9660_xa_t *xa_data = NULL;
   uint8_t dir_len= iso9660_get_dir_len(p_iso9660_dir);
-  unsigned int filename_len;
+  iso711_t i_fname;
   unsigned int stat_len;
-  iso9660_stat_t *stat;
+  iso9660_stat_t *p_stat;
 
   if (!dir_len) return NULL;
 
-  filename_len  = from_711(p_iso9660_dir->filename_len);
+  i_fname  = from_711(p_iso9660_dir->filename_len);
 
   /* .. string in statbuf is one longer than in p_iso9660_dir's listing '\1' */
-  stat_len      = sizeof(iso9660_stat_t)+filename_len+2;
+  stat_len      = sizeof(iso9660_stat_t)+i_fname+2;
 
-  stat          = _cdio_malloc(stat_len);
-  stat->type    = (p_iso9660_dir->file_flags & ISO_DIRECTORY) 
+  p_stat          = calloc(1, stat_len);
+  p_stat->type    = (p_iso9660_dir->file_flags & ISO_DIRECTORY) 
     ? _STAT_DIR : _STAT_FILE;
-  stat->lsn     = from_733 (p_iso9660_dir->extent);
-  stat->size    = from_733 (p_iso9660_dir->size);
-  stat->secsize = _cdio_len2blocks (stat->size, ISO_BLOCKSIZE);
+  p_stat->lsn     = from_733 (p_iso9660_dir->extent);
+  p_stat->size    = from_733 (p_iso9660_dir->size);
+  p_stat->secsize = _cdio_len2blocks (p_stat->size, ISO_BLOCKSIZE);
+  p_stat->rr.b3_rock = dunno; /*FIXME should do based on mask */
 
-  if ('\0' == p_iso9660_dir->filename[0] && 1 == filename_len)
-    strcpy (stat->filename, ".");
-  else if ('\1' == p_iso9660_dir->filename[0] && 1 == filename_len)
-    strcpy (stat->filename, "..");
-  else {
+  {
+    char rr_fname[256] = "";
+    int  i_rr_fname = 
+      get_rock_ridge_filename(p_iso9660_dir, rr_fname, p_stat);
+    if (i_rr_fname > 0) {
+      if (i_rr_fname > i_fname) {
+	/* realloc gives valgrind errors */
+	iso9660_stat_t *p_stat_new = 
+	  calloc(1, sizeof(iso9660_stat_t)+i_rr_fname+2);
+	memcpy(p_stat_new, p_stat, stat_len);
+	free(p_stat);
+	p_stat = p_stat_new;
+      }
+      strncpy(p_stat->filename, rr_fname, i_rr_fname+1);
+    } else {
+      if ('\0' == p_iso9660_dir->filename[0] && 1 == i_fname)
+	strcpy (p_stat->filename, ".");
+      else if ('\1' == p_iso9660_dir->filename[0] && 1 == i_fname)
+	strcpy (p_stat->filename, "..");
 #ifdef HAVE_JOLIET
-    if (i_joliet_level) {
-      int i_inlen = filename_len;
-      int i_outlen = (i_inlen / 2);
-      char *p_psz_out = NULL;
-      ucs2be_to_locale(p_iso9660_dir->filename, i_inlen, 
-		       &p_psz_out, i_outlen);
-      strncpy(stat->filename, p_psz_out, filename_len);
-      free(p_psz_out);
-    } else
+      else if (i_joliet_level) {
+	int i_inlen = i_fname;
+	int i_outlen = (i_inlen / 2);
+	char *p_psz_out = NULL;
+	ucs2be_to_locale(p_iso9660_dir->filename, i_inlen, &p_psz_out, 
+			 i_outlen);
+	strncpy(p_stat->filename, p_psz_out, i_fname);
+	free(p_psz_out);
+      }
 #endif /*HAVE_JOLIET*/
-      strncpy (stat->filename, p_iso9660_dir->filename, filename_len);
+      else {
+	strncpy (p_stat->filename, p_iso9660_dir->filename, i_fname);
+      }
+    }
   }
   
 
-  iso9660_get_dtime(&(p_iso9660_dir->recording_time), true, &(stat->tm));
+  iso9660_get_dtime(&(p_iso9660_dir->recording_time), true, &(p_stat->tm));
 
-  cdio_assert (dir_len >= sizeof (iso9660_dir_t));
+  if (dir_len < sizeof (iso9660_dir_t)) {
+    free(p_stat->rr.psz_symlink);
+    free(p_stat);
+    return NULL;
+  }
+  
 
-  if (b_mode2) {
+  {
     int su_length = iso9660_get_dir_len(p_iso9660_dir) 
       - sizeof (iso9660_dir_t);
-    su_length -= filename_len;
+    su_length -= i_fname;
     
     if (su_length % 2)
       su_length--;
     
     if (su_length < 0 || su_length < sizeof (iso9660_xa_t))
-      return stat;
+      return p_stat;
     
-    xa_data = (void *) (((char *) p_iso9660_dir) 
-			+ (iso9660_get_dir_len(p_iso9660_dir) - su_length));
-    
-    if (xa_data->signature[0] != 'X' 
-	|| xa_data->signature[1] != 'A')
-    {
-      cdio_warn ("XA signature not found in ISO9660's system use area;"
-		 " ignoring XA attributes for this file entry.");
-      cdio_debug ("%d %d %d, '%c%c' (%d, %d)", 
-		  iso9660_get_dir_len(p_iso9660_dir), 
-		  filename_len,
-		  su_length,
-		  xa_data->signature[0], xa_data->signature[1],
-		  xa_data->signature[0], xa_data->signature[1]);
-      return stat;
+    if (nope == b_xa) {
+      return p_stat;
+    } else {
+      iso9660_xa_t *xa_data = 
+	(void *) (((char *) p_iso9660_dir)  
+		  + (iso9660_get_dir_len(p_iso9660_dir) - su_length));
+      cdio_log_level_t loglevel = (yep == b_xa) 
+	? CDIO_LOG_WARN : CDIO_LOG_INFO;
+      
+      if (xa_data->signature[0] != 'X' 
+	  || xa_data->signature[1] != 'A')
+	{
+	  cdio_log (loglevel, 
+		    "XA signature not found in ISO9660's system use area;"
+		     " ignoring XA attributes for this file entry.");
+	  cdio_debug ("%d %d %d, '%c%c' (%d, %d)", 
+		      iso9660_get_dir_len(p_iso9660_dir), 
+		      i_fname,
+		      su_length,
+		      xa_data->signature[0], xa_data->signature[1],
+		      xa_data->signature[0], xa_data->signature[1]);
+	  return p_stat;
+	}
+      p_stat->xa = *xa_data;
     }
-    stat->xa = *xa_data;
   }
-  return stat;
+  return p_stat;
     
 }
 
@@ -713,9 +939,9 @@ _fs_stat_root (CdIo_t *p_cdio)
   {
     iso_extension_mask_t iso_extension_mask = ISO_EXTENSION_ALL;
     generic_img_private_t *p_env = (generic_img_private_t *) p_cdio->env;
-    bool b_mode2 = cdio_get_track_green(p_cdio, 1);
     iso9660_dir_t *p_iso9660_dir;
     iso9660_stat_t *p_stat;
+    bool_3way_t b_xa;
 
     if (!p_env->i_joliet_level)
       iso_extension_mask &= ~ISO_EXTENSION_JOLIET;
@@ -726,6 +952,17 @@ _fs_stat_root (CdIo_t *p_cdio)
       return NULL;
     }
 
+    switch(cdio_get_discmode(p_cdio)) {
+    case CDIO_DISC_MODE_CD_XA: 
+      b_xa = yep;
+      break;
+    case CDIO_DISC_MODE_CD_DATA: 
+      b_xa = nope;
+      break;
+    default: 
+      b_xa = dunno;
+    }
+
 #ifdef HAVE_JOLIET    
     p_iso9660_dir = p_env->i_joliet_level 
       ? &(p_env->svd.root_directory_record) 
@@ -734,7 +971,7 @@ _fs_stat_root (CdIo_t *p_cdio)
     p_iso9660_dir = &(p_env->pvd.root_directory_record) ;
 #endif
     
-    p_stat = _iso9660_dir_to_statbuf (p_iso9660_dir, b_mode2, 
+    p_stat = _iso9660_dir_to_statbuf (p_iso9660_dir, b_xa, 
 				      p_env->i_joliet_level);
     return p_stat;
   }
@@ -755,14 +992,14 @@ _fs_stat_iso_root (iso9660_t *p_iso)
   p_iso9660_dir = &(p_iso->pvd.root_directory_record) ;
 #endif
   
-  p_stat = _iso9660_dir_to_statbuf (p_iso9660_dir, true, 
+  p_stat = _iso9660_dir_to_statbuf (p_iso9660_dir, p_iso->b_xa,
 				    p_iso->i_joliet_level);
   return p_stat;
 }
 
 static iso9660_stat_t *
 _fs_stat_traverse (const CdIo_t *p_cdio, const iso9660_stat_t *_root, 
-		   char **splitpath, bool b_mode2, bool translate)
+		   char **splitpath)
 {
   unsigned offset = 0;
   uint8_t *_dirbuf = NULL;
@@ -772,7 +1009,7 @@ _fs_stat_traverse (const CdIo_t *p_cdio, const iso9660_stat_t *_root,
   if (!splitpath[0])
     {
       unsigned int len=sizeof(iso9660_stat_t) + strlen(_root->filename)+1;
-      p_stat = _cdio_malloc(len);
+      p_stat = calloc(1, len);
       memcpy(p_stat, _root, len);
       return p_stat;
     }
@@ -789,17 +1026,11 @@ _fs_stat_traverse (const CdIo_t *p_cdio, const iso9660_stat_t *_root,
 		 (unsigned long int) ISO_BLOCKSIZE * _root->secsize);
     }
   
-  _dirbuf = _cdio_malloc (_root->secsize * ISO_BLOCKSIZE);
+  _dirbuf = calloc(1, _root->secsize * ISO_BLOCKSIZE);
 
-  if (b_mode2) {
-    if (cdio_read_mode2_sectors (p_cdio, _dirbuf, _root->lsn, false, 
-				 _root->secsize))
+  if (cdio_read_data_sectors (p_cdio, _dirbuf, _root->lsn, ISO_BLOCKSIZE, 
+			      _root->secsize))
       return NULL;
-  } else {
-    if (cdio_read_mode1_sectors (p_cdio, _dirbuf, _root->lsn, false,
-				 _root->secsize))
-      return NULL;
-  }
   
   while (offset < (_root->secsize * ISO_BLOCKSIZE))
     {
@@ -813,35 +1044,40 @@ _fs_stat_traverse (const CdIo_t *p_cdio, const iso9660_stat_t *_root,
 	  continue;
 	}
       
-      p_stat = _iso9660_dir_to_statbuf (p_iso9660_dir, b_mode2, 
+      p_stat = _iso9660_dir_to_statbuf (p_iso9660_dir, dunno, 
 					p_env->i_joliet_level);
 
-      if (translate) {
-	char *trans_fname = malloc(strlen(p_stat->filename));
+      cmp = strcmp(splitpath[0], p_stat->filename);
+
+      if ( 0 != cmp && 0 == p_env->i_joliet_level && yep != p_stat->rr.b3_rock ) {
+	char *trans_fname = NULL;
+	unsigned int i_trans_fname=strlen(p_stat->filename);
 	int trans_len;
 	
-	if (trans_fname == NULL) {
-	  cdio_warn("can't allocate %lu bytes", 
-		    (long unsigned int) strlen(p_stat->filename));
-	  return NULL;
+	if (i_trans_fname) {
+	  trans_fname = calloc(1, i_trans_fname+1);
+	  if (!trans_fname) {
+	    cdio_warn("can't allocate %lu bytes", 
+		      (long unsigned int) strlen(p_stat->filename));
+	    return NULL;
+	  }
+	  trans_len = iso9660_name_translate_ext(p_stat->filename, trans_fname,
+						 p_env->i_joliet_level);
+	  cmp = strcmp(splitpath[0], trans_fname);
+	  free(trans_fname);
 	}
-	trans_len = iso9660_name_translate_ext(p_stat->filename, trans_fname, 
-					       p_env->i_joliet_level);
-	cmp = strcmp(splitpath[0], trans_fname);
-	free(trans_fname);
-      } else {
-	cmp = strcmp(splitpath[0], p_stat->filename);
       }
       
       if (!cmp) {
 	iso9660_stat_t *ret_stat 
-	  = _fs_stat_traverse (p_cdio, p_stat, &splitpath[1], b_mode2, 
-			       translate);
+	  = _fs_stat_traverse (p_cdio, p_stat, &splitpath[1]);
+	free(p_stat->rr.psz_symlink);
 	free(p_stat);
 	free (_dirbuf);
 	return ret_stat;
       }
 
+      free(p_stat->rr.psz_symlink);
       free(p_stat);
 	  
       offset += iso9660_get_dir_len(p_iso9660_dir);
@@ -856,7 +1092,7 @@ _fs_stat_traverse (const CdIo_t *p_cdio, const iso9660_stat_t *_root,
 
 static iso9660_stat_t *
 _fs_iso_stat_traverse (iso9660_t *p_iso, const iso9660_stat_t *_root, 
-		       char **splitpath, bool translate)
+		       char **splitpath)
 {
   unsigned offset = 0;
   uint8_t *_dirbuf = NULL;
@@ -866,7 +1102,7 @@ _fs_iso_stat_traverse (iso9660_t *p_iso, const iso9660_stat_t *_root,
     {
       iso9660_stat_t *p_stat;
       unsigned int len=sizeof(iso9660_stat_t) + strlen(_root->filename)+1;
-      p_stat = _cdio_malloc(len);
+      p_stat = calloc(1, len);
       memcpy(p_stat, _root, len);
       return p_stat;
     }
@@ -883,7 +1119,7 @@ _fs_iso_stat_traverse (iso9660_t *p_iso, const iso9660_stat_t *_root,
 		 (unsigned long int) ISO_BLOCKSIZE * _root->secsize);
     }
   
-  _dirbuf = _cdio_malloc (_root->secsize * ISO_BLOCKSIZE);
+  _dirbuf = calloc(1, _root->secsize * ISO_BLOCKSIZE);
 
   ret = iso9660_iso_seek_read (p_iso, _dirbuf, _root->lsn, _root->secsize);
   if (ret!=ISO_BLOCKSIZE*_root->secsize) return NULL;
@@ -900,10 +1136,13 @@ _fs_iso_stat_traverse (iso9660_t *p_iso, const iso9660_stat_t *_root,
 	  continue;
 	}
       
-      p_stat = _iso9660_dir_to_statbuf (p_iso9660_dir, true, 
+      p_stat = _iso9660_dir_to_statbuf (p_iso9660_dir, p_iso->b_xa, 
 					p_iso->i_joliet_level);
 
-      if (translate) {
+      cmp = strcmp(splitpath[0], p_stat->filename);
+
+      if ( 0 != cmp && 0 == p_iso->i_joliet_level 
+	   && yep != p_stat->rr.b3_rock ) {
 	char *trans_fname = malloc(strlen(p_stat->filename)+1);
 	int trans_len;
 	
@@ -916,18 +1155,18 @@ _fs_iso_stat_traverse (iso9660_t *p_iso, const iso9660_stat_t *_root,
 					       p_iso->i_joliet_level);
 	cmp = strcmp(splitpath[0], trans_fname);
 	free(trans_fname);
-      } else {
-	cmp = strcmp(splitpath[0], p_stat->filename);
       }
       
       if (!cmp) {
 	iso9660_stat_t *ret_stat 
-	  = _fs_iso_stat_traverse (p_iso, p_stat, &splitpath[1], translate);
+	  = _fs_iso_stat_traverse (p_iso, p_stat, &splitpath[1]);
+	free(p_stat->rr.psz_symlink);
 	free(p_stat);
 	free (_dirbuf);
 	return ret_stat;
       }
 
+      free(p_stat->rr.psz_symlink);
       free(p_stat);
 	  
       offset += iso9660_get_dir_len(p_iso9660_dir);
@@ -944,23 +1183,21 @@ _fs_iso_stat_traverse (iso9660_t *p_iso, const iso9660_stat_t *_root,
   Get file status for pathname into stat. NULL is returned on error.
  */
 iso9660_stat_t *
-iso9660_fs_stat (CdIo_t *p_cdio, const char pathname[])
+iso9660_fs_stat (CdIo_t *p_cdio, const char psz_path[])
 {
   iso9660_stat_t *p_root;
   char **p_psz_splitpath;
   iso9660_stat_t *p_stat;
-  /* A bit of a hack, we'll assume track 1 contains ISO_PVD_SECTOR.*/
-  bool b_mode2;
 
   if (!p_cdio)   return NULL;
-  if (!pathname) return NULL;
+  if (!psz_path) return NULL;
 
   p_root = _fs_stat_root (p_cdio);
-  if (!p_root) return NULL;
 
-  b_mode2 = cdio_get_track_green(p_cdio, 1);
-  p_psz_splitpath = _cdio_strsplit (pathname, '/');
-  p_stat = _fs_stat_traverse (p_cdio, p_root, p_psz_splitpath, b_mode2, false);
+  if (!p_root)   return NULL;
+
+  p_psz_splitpath = _cdio_strsplit (psz_path, '/');
+  p_stat = _fs_stat_traverse (p_cdio, p_root, p_psz_splitpath);
   free(p_root);
   _cdio_strfreev (p_psz_splitpath);
 
@@ -974,7 +1211,7 @@ iso9660_fs_stat (CdIo_t *p_cdio, const char pathname[])
   are lowercased.
  */
 iso9660_stat_t *
-iso9660_fs_stat_translate (CdIo_t *p_cdio, const char pathname[], 
+iso9660_fs_stat_translate (CdIo_t *p_cdio, const char psz_path[], 
 			   bool b_mode2)
 {
   iso9660_stat_t *p_root;
@@ -982,13 +1219,13 @@ iso9660_fs_stat_translate (CdIo_t *p_cdio, const char pathname[],
   iso9660_stat_t *p_stat;
 
   if (!p_cdio)  return NULL;
-  if (pathname) return NULL;
+  if (psz_path) return NULL;
 
   p_root = _fs_stat_root (p_cdio);
   if (!p_root) return NULL;
 
-  p_psz_splitpath = _cdio_strsplit (pathname, '/');
-  p_stat = _fs_stat_traverse (p_cdio, p_root, p_psz_splitpath, b_mode2, true);
+  p_psz_splitpath = _cdio_strsplit (psz_path, '/');
+  p_stat = _fs_stat_traverse (p_cdio, p_root, p_psz_splitpath);
   free(p_root);
   _cdio_strfreev (p_psz_splitpath);
 
@@ -999,22 +1236,22 @@ iso9660_fs_stat_translate (CdIo_t *p_cdio, const char pathname[],
   Get file status for pathname into stat. NULL is returned on error.
  */
 iso9660_stat_t *
-iso9660_ifs_stat (iso9660_t *p_iso, const char pathname[])
+iso9660_ifs_stat (iso9660_t *p_iso, const char psz_path[])
 {
   iso9660_stat_t *p_root;
   char **splitpath;
   iso9660_stat_t *stat;
 
   if (!p_iso)    return NULL;
-  if (!pathname) return NULL;
+  if (!psz_path) return NULL;
 
   p_root = _fs_stat_iso_root (p_iso);
   if (!p_root) return NULL;
 
-  splitpath = _cdio_strsplit (pathname, '/');
-  stat = _fs_iso_stat_traverse (p_iso, p_root, splitpath, false);
+  splitpath = _cdio_strsplit (psz_path, '/');
+  stat = _fs_iso_stat_traverse (p_iso, p_root, splitpath);
   free(p_root);
-  /*** FIXME _cdio_strfreev (splitpath); ***/
+  _cdio_strfreev (splitpath);
 
   return stat;
 }
@@ -1026,20 +1263,20 @@ iso9660_ifs_stat (iso9660_t *p_iso, const char pathname[])
   are lowercased.
  */
 iso9660_stat_t *
-iso9660_ifs_stat_translate (iso9660_t *p_iso, const char pathname[])
+iso9660_ifs_stat_translate (iso9660_t *p_iso, const char psz_path[])
 {
   iso9660_stat_t *p_root;
   char **p_psz_splitpath;
   iso9660_stat_t *p_stat;
 
   if (!p_iso)    return NULL;
-  if (!pathname) return NULL;
+  if (!psz_path) return NULL;
 
   p_root = _fs_stat_iso_root (p_iso);
   if (NULL == p_root) return NULL;
 
-  p_psz_splitpath = _cdio_strsplit (pathname, '/');
-  p_stat = _fs_iso_stat_traverse (p_iso, p_root, p_psz_splitpath, true);
+  p_psz_splitpath = _cdio_strsplit (psz_path, '/');
+  p_stat = _fs_iso_stat_traverse (p_iso, p_root, p_psz_splitpath);
   free(p_root);
   _cdio_strfreev (p_psz_splitpath);
 
@@ -1051,18 +1288,19 @@ iso9660_ifs_stat_translate (iso9660_t *p_iso, const char pathname[])
   of the files inside that. The caller must free the returned result.
 */
 CdioList_t * 
-iso9660_fs_readdir (CdIo_t *p_cdio, const char pathname[], bool b_mode2)
+iso9660_fs_readdir (CdIo_t *p_cdio, const char psz_path[], bool b_mode2)
 {
   iso9660_stat_t *p_stat;
   generic_img_private_t *p_env = (generic_img_private_t *) p_cdio->env;
 
   if (!p_cdio)   return NULL;
-  if (!pathname) return NULL;
+  if (!psz_path) return NULL;
 
-  p_stat = iso9660_fs_stat (p_cdio, pathname);
+  p_stat = iso9660_fs_stat (p_cdio, psz_path);
   if (!p_stat) return NULL;
 
   if (p_stat->type != _STAT_DIR) {
+    free(p_stat->rr.psz_symlink);
     free(p_stat);
     return NULL;
   }
@@ -1079,17 +1317,11 @@ iso9660_fs_readdir (CdIo_t *p_cdio, const char pathname[], bool b_mode2)
 		   (unsigned long int) ISO_BLOCKSIZE * p_stat->secsize);
       }
 
-    _dirbuf = _cdio_malloc (p_stat->secsize * ISO_BLOCKSIZE);
+    _dirbuf = calloc(1, p_stat->secsize * ISO_BLOCKSIZE);
 
-    if (b_mode2) {
-      if (cdio_read_mode2_sectors (p_cdio, _dirbuf, p_stat->lsn, false, 
-				   p_stat->secsize))
-	cdio_assert_not_reached ();
-    } else {
-      if (cdio_read_mode1_sectors (p_cdio, _dirbuf, p_stat->lsn, false,
-				   p_stat->secsize))
-	cdio_assert_not_reached ();
-    }
+    if (cdio_read_data_sectors (p_cdio, _dirbuf, p_stat->lsn, 
+				ISO_BLOCKSIZE, p_stat->secsize))
+	return NULL;
 
     while (offset < (p_stat->secsize * ISO_BLOCKSIZE))
       {
@@ -1102,7 +1334,7 @@ iso9660_fs_readdir (CdIo_t *p_cdio, const char pathname[], bool b_mode2)
 	    continue;
 	  }
 
-	p_iso9660_stat = _iso9660_dir_to_statbuf(p_iso9660_dir, b_mode2, 
+	p_iso9660_stat = _iso9660_dir_to_statbuf(p_iso9660_dir, dunno,
 						 p_env->i_joliet_level);
 	_cdio_list_append (retval, p_iso9660_stat);
 
@@ -1122,17 +1354,18 @@ iso9660_fs_readdir (CdIo_t *p_cdio, const char pathname[], bool b_mode2)
   of the files inside that. The caller must free the returned result.
 */
 CdioList_t * 
-iso9660_ifs_readdir (iso9660_t *p_iso, const char pathname[])
+iso9660_ifs_readdir (iso9660_t *p_iso, const char psz_path[])
 {
   iso9660_stat_t *p_stat;
 
   if (!p_iso)    return NULL;
-  if (!pathname) return NULL;
+  if (!psz_path) return NULL;
 
-  p_stat = iso9660_ifs_stat (p_iso, pathname);
+  p_stat = iso9660_ifs_stat (p_iso, psz_path);
   if (!p_stat)   return NULL;
 
   if (p_stat->type != _STAT_DIR) {
+    free(p_stat->rr.psz_symlink);
     free(p_stat);
     return NULL;
   }
@@ -1150,7 +1383,7 @@ iso9660_ifs_readdir (iso9660_t *p_iso, const char pathname[])
 		   (unsigned long int) ISO_BLOCKSIZE * p_stat->secsize);
       }
 
-    _dirbuf = _cdio_malloc (p_stat->secsize * ISO_BLOCKSIZE);
+    _dirbuf = calloc(1, p_stat->secsize * ISO_BLOCKSIZE);
 
     ret = iso9660_iso_seek_read (p_iso, _dirbuf, p_stat->lsn, p_stat->secsize);
     if (ret != ISO_BLOCKSIZE*p_stat->secsize) return NULL;
@@ -1166,16 +1399,23 @@ iso9660_ifs_readdir (iso9660_t *p_iso, const char pathname[])
 	    continue;
 	  }
 
-	p_iso9660_stat = _iso9660_dir_to_statbuf(p_iso9660_dir, true,
+	p_iso9660_stat = _iso9660_dir_to_statbuf(p_iso9660_dir, p_iso->b_xa,
 						 p_iso->i_joliet_level);
-	_cdio_list_append (retval, p_iso9660_stat);
+
+	if (p_iso9660_stat) 
+	  _cdio_list_append (retval, p_iso9660_stat);
 
 	offset += iso9660_get_dir_len(p_iso9660_dir);
       }
 
-    cdio_assert (offset == (p_stat->secsize * ISO_BLOCKSIZE));
-
     free (_dirbuf);
+
+    if (offset != (p_stat->secsize * ISO_BLOCKSIZE)) {
+      free (p_stat);
+      _cdio_list_free (retval, true);
+      return NULL;
+    }
+
     free (p_stat);
     return retval;
   }
@@ -1209,7 +1449,7 @@ find_fs_lsn_recurse (CdIo_t *p_cdio, const char pathname[], lsn_t lsn)
 
       if (statbuf->lsn == lsn) {
 	unsigned int len=sizeof(iso9660_stat_t)+strlen(statbuf->filename)+1;
-	iso9660_stat_t *ret_stat = _cdio_malloc(len);
+	iso9660_stat_t *ret_stat = calloc(1, len);
 	memcpy(ret_stat, statbuf, len);
         _cdio_list_free (entlist, true);
         _cdio_list_free (dirlist, true);
@@ -1256,5 +1496,5 @@ bool
 iso9660_ifs_is_xa (const iso9660_t * p_iso) 
 {
   if (!p_iso) return false;
-  return p_iso->b_xa;
+  return yep == p_iso->b_xa;
 }
