@@ -1,8 +1,8 @@
 /*
-    $Id: xa.c,v 1.1 2004/12/18 17:29:32 rocky Exp $
+    $Id: xa.c,v 1.4 2005/02/20 10:21:01 rocky Exp $
 
     Copyright (C) 2000 Herbert Valerio Riedel <hvr@gnu.org>
-    Copyright (C) 2003 Rocky Bernstein <rocky@panix.com>
+    Copyright (C) 2003, 2005 Rocky Bernstein <rocky@panix.com>
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -28,6 +28,12 @@
 #include <string.h>
 #endif
 
+/*! String inside frame which identifies XA attributes.  Note should
+    come *before* public headers which does a #define of
+    this name.
+*/
+const char ISO_XA_MARKER_STRING[] = {'C', 'D', '-', 'X', 'A', '0', '0', '1'};
+
 /* Public headers */
 #include <cdio/iso9660.h>
 #include <cdio/util.h>
@@ -36,6 +42,9 @@
 /* Private headers */
 #include "cdio_assert.h"
 
+/* Variables to hold debugger-helping enumerations */
+enum cdio_xa_enums;
+
 #define BUF_COUNT 16
 #define BUF_SIZE 80
 
